#
# coding=utf-8
"""template - Template utilities for hubugs"""
# Copyright (C) 2010-2012  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import datetime
import operator
import os
import sys
import subprocess
import tempfile

import html2text as html2
import jinja2

from pygments import highlight as pyg_highlight
from pygments.formatters import get_formatter_by_name
from pygments.lexers import get_lexer_by_name

from . import utils


if sys.platform == 'darwin':
    _USER_DATA_DIR = os.path.expanduser('~/Library/Application Support')
else:
    _USER_DATA_DIR = os.path.join(os.environ.get("HOME", "/"), ".local")

USER_DATA_DIR = os.environ.get("XDG_DATA_HOME", _USER_DATA_DIR)
SYSTEM_DATA_DIR = os.environ.get("XDG_DATA_DIRS",
                                 "/usr/local/share/:/usr/share/").split(":")
PKG_DATA_DIRS = [os.path.join(USER_DATA_DIR, "hubugs", "templates"), ]
for directory in SYSTEM_DATA_DIR:
    PKG_DATA_DIRS.append(os.path.join(directory, "hubugs", "templates"))

ENV = jinja2.Environment(loader=jinja2.ChoiceLoader(
    [jinja2.FileSystemLoader(s) for s in PKG_DATA_DIRS]))
ENV.loader.loaders.append(jinja2.PackageLoader("hubugs", "templates"))


class EmptyMessageError(ValueError):

    """Error to raise when the user provides an empty message."""

    pass


def get_template(group, name):
    """Fetch a Jinja template instance.

    :param str group: Template group identifier
    :param str name: Template name
    :rtype: ``jinja2.environment.Template``
    :return: Jinja template instance

    """
    template_set = utils.get_git_config_val('hubugs.templates', 'default')
    return ENV.get_template("%s/%s/%s" % (template_set, group, name))


def jinja_filter(func):
    """Simple decorator to add a new filter to Jinja environment.

    :param func func: Function to add to Jinja environment
    :rtype: ``func``
    :returns: Unmodified function

    """
    ENV.filters[func.__name__] = func

    return func


@jinja_filter
def colourise(text, formatting):
    """Colourise text using blessings.

    Returns text untouched if colour output is not enabled

    :see: ``blessings``

    :param str text: Text to colourise
    :param str formatting: Formatting to apply to text
    :rtype: ``str``
    :return: Colourised text, when possible

    """
    return getattr(utils.T, formatting.replace(' ', '_'))(text)
# American spelling, just for Brandon Cady ;)
ENV.filters["colorize"] = ENV.filters["colourise"]


@jinja_filter
def highlight(text, lexer="diff", formatter="terminal"):
    """Highlight text with pygments.

    Returns text untouched if colour output is not enabled

    :param str text: Text to highlight
    :param str lexer: Jinja lexer to use
    :param str formatter: Jinja formatter to use
    :rtype: ``str``
    :return: Syntax highlighted output, when possible

    """
    if utils.T.is_a_tty:
        lexer = get_lexer_by_name(lexer)
        formatter = get_formatter_by_name(formatter)
        return pyg_highlight(text, lexer, formatter)
    else:
        return text


@jinja_filter
def html2text(html, width=80, ascii_replacements=False):
    """HTML to plain text renderer.

    :param str text: Text to process
    :param int width: Paragraph width
    :param bool ascii_replacements: Use psuedo-ascii replacements for Unicode
    :rtype: ``str``
    :return: Rendered text

    """
    html2.BODY_WIDTH = width
    html2.UNICODE_SNOB = ascii_replacements
    return html2.html2text(html).strip()


@jinja_filter
def relative_time(timestamp):
    """Format a relative time.

    Taken from bleeter_.  Duplication is evil, I know.

    :param datetime.datetime timestamp: Event to generate relative timestamp
        against
    :rtype: ``str``
    :return: Human readable date and time offset

    .. _bleeter: http://jnrowe.github.com/bleeter/

    """

    numstr = ". a two three four five six seven eight nine ten".split()

    matches = [
        60 * 60 * 24 * 365,
        60 * 60 * 24 * 28,
        60 * 60 * 24 * 7,
        60 * 60 * 24,
        60 * 60,
        60,
        1,
    ]
    match_names = ["year", "month", "week", "day", "hour", "minute", "second"]

    delta = datetime.datetime.utcnow() - timestamp
    # Switch to delta.total_seconds, if 2.6 support is dropped
    seconds = delta.days * 86400 + delta.seconds
    for scale in matches:
        i = seconds // scale
        if i:
            name = match_names[matches.index(scale)]
            break

    if i == 1 and name in ("year", "month", "week"):
        result = "last %s" % name
    elif i == 1 and name == "day":
        result = "yesterday"
    elif i == 1 and name == "hour":
        result = "about an hour ago"
    else:
        result = "about %s %s%s ago" % (i if i > 10 else numstr[i], name,
                                        "s" if i > 1 else "")
    return result


def display_bugs(bugs, order, **extras):
    """Display bugs to users.

    :type bugs: ``list` of ``models.Issue``
    :param bugs: Bugs to display
    :param str order: Sorting order for displaying bugs
    :param dict extras: Additional values to pass to templates
    :rtype: ``str``
    :return: Rendered template output

    """
    if not bugs:
        return utils.success("No bugs found!")

    # Match ordering method to bug attribute
    if order == "updated":
        attr = "updated_at"
    else:
        attr = order

    bugs = sorted(bugs, key=operator.attrgetter(attr))

    # Default to 80 columns, when stdout is not a tty
    columns = utils.T.width if utils.T.width else 80

    template = get_template('view', 'list.txt')

    max_id = max(i.number for i in bugs)
    id_len = len(str(max_id))
    spacer = " " * (id_len - 2)

    return template.render(bugs=bugs, spacer=spacer, id_len=id_len,
                           max_title=columns - id_len - 2, **extras)


def edit_text(edit_type="default", data=None):
    """Edit data with external editor.

    :param str edit_type: Template to use in editor
    :param dict data: Information to pass to template
    :rtype: ``str``
    :return: User supplied text
    :raise EmptyMessageError: No message given

    """
    template = get_template('edit', '%s.mkd' % edit_type)
    with tempfile.NamedTemporaryFile(prefix="hubugs-", suffix=".mkd") as temp:
        temp.write(template.render(data if data else {}))
        temp.flush()
        orig_mtime = os.path.getmtime(temp.name)

        subprocess.check_call([utils.get_editor(), temp.name])

        temp.seek(0)
        text = "".join(filter(lambda s: not s.startswith("#"),
                              temp.readlines())).strip()

        if not text:
            raise EmptyMessageError("No message given")
        elif orig_mtime == os.path.getmtime(temp.name):
            raise EmptyMessageError("Message not edited")

    return text.strip()



import argparse

from subprocess import CalledProcessError
from unittest import TestCase

from mock import (Mock, patch)
from nose.tools import (assert_equals, raises)

from hubugs import utils


# We only test forced styling output of blessings, as blessings handles the
# sys.stdout.isatty() flipping
utils.T = utils.blessings.Terminal(force_styling=True)


def fake_env(key, default=None):
    """Fake environment settings used for os.getenv mocking"""
    fake_data = {
        'HUBUGS_TOKEN': 'xxx',
        'HOME': '/home/JNRowe',
        'XDG_CACHE_HOME': 'cache_dir',
    }
    return fake_data[key]


def test_colouriser():
    assert_equals(utils.success('test'), u'\x1b[38;5;10mtest\x1b[m\x1b(B')
    assert_equals(utils.fail('test'), u'\x1b[38;5;9mtest\x1b[m\x1b(B')
    assert_equals(utils.warn('test'), u'\x1b[38;5;11mtest\x1b[m\x1b(B')


class ProjectAction(TestCase):
    def setUp(self):
        self.parser = utils.argh.ArghParser('test_parser')
        # Stub out ._print_message() to stop help messages being displayed on
        # stderr during tests.
        self.parser._print_message = Mock(return_value=True)
        self.namespace = argparse.Namespace()
        self.action = utils.ProjectAction([], '')

    @patch('hubugs.utils.get_github_api')
    @patch('hubugs.utils.get_git_config_val')
    def test_repo_name(self, get_git_config_val, get_github_api):
        get_github_api().repos.show = Mock(return_value=True)
        get_git_config_val.return_value = 'JNRowe'

        self.action(self.parser, self.namespace, 'misc-overlay')
        assert_equals(self.namespace.project, 'JNRowe/misc-overlay')

        self.action(self.parser, self.namespace, 'JNRowe/misc-overlay')
        assert_equals(self.namespace.project, 'JNRowe/misc-overlay')

        self.action(self.parser, self.namespace, 'ask/python-github2')
        assert_equals(self.namespace.project, 'ask/python-github2')

    @patch('hubugs.utils.get_github_api')
    @patch('hubugs.utils.get_git_config_val')
    @raises(SystemExit)
    def test_no_user(self, get_git_config_val, get_github_api):
        get_github_api().repos.show = Mock(return_value=True)
        get_git_config_val.return_value = None
        self.action(self.parser, self.namespace, 'misc-overlay')


class GetGitConfigVal(TestCase):
    @patch('hubugs.utils.check_output')
    def test_valid_key(self, check_output):
        check_output.return_value = 'JNRowe'
        assert_equals(utils.get_git_config_val('github.user'), 'JNRowe')

    @patch('hubugs.utils.check_output')
    def test_invalid_key(self, check_output):
        check_output.return_value = ''
        assert_equals(utils.get_git_config_val('no_such_key'), '')

    @patch('hubugs.utils.check_output')
    def test_command_error(self, check_output):
        check_output.side_effect = CalledProcessError('255', 'cmd')
        assert_equals(utils.get_git_config_val('github.user'), None)


class GetEditor(TestCase):
    @patch('os.getenv')
    def test_git_editor_envvar(self, getenv):
        getenv.return_value = 'custom git editor'
        assert_equals(utils.get_editor(), 'custom git editor')

    @patch('hubugs.utils.get_git_config_val')
    @patch('os.getenv')
    def test_git_editor_config(self, getenv, get_git_config_val):
        getenv.return_value = None
        get_git_config_val.return_value = 'custom config editor'
        assert_equals(utils.get_editor(), 'custom config editor')

    @patch('hubugs.utils.get_git_config_val')
    @patch('os.getenv')
    def test_editor_environment_visual(self, getenv, get_git_config_val):
        def fake_env(key, default=None):
            return {'VISUAL': 'visual'}.get(key)
        getenv.side_effect = fake_env
        get_git_config_val.return_value = None
        assert_equals(utils.get_editor(), 'visual')

    @patch('hubugs.utils.get_git_config_val')
    @patch('os.getenv')
    def test_editor_environment_editor(self, getenv, get_git_config_val):
        def fake_env(key, default=None):
            return {'EDITOR': 'editor'}.get(key, default)
        getenv.side_effect = fake_env
        get_git_config_val.return_value = None
        assert_equals(utils.get_editor(), 'editor')

    @patch('hubugs.utils.get_git_config_val')
    @patch('os.getenv')
    def test_editor_default(self, getenv, get_git_config_val):
        def fake_env(key, default=None):
            return default
        getenv.side_effect = fake_env
        get_git_config_val.return_value = None
        assert_equals(utils.get_editor(), 'vi')


class GetRepo(TestCase):
    @patch('hubugs.utils.get_git_config_val')
    def test_ssh_url(self, get_git_config_val):
        get_git_config_val.return_value = \
            'git@github.com:JNRowe/misc-overlay.git'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_ssh_url_no_suffix(self, get_git_config_val):
        get_git_config_val.return_value = \
            'git@github.com:JNRowe/misc-overlay'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_git_url(self, get_git_config_val):
        get_git_config_val.return_value = \
            'git://github.com/JNRowe/misc-overlay.git'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_git_url_no_suffix(self, get_git_config_val):
        get_git_config_val.return_value = \
            'git://github.com/JNRowe/misc-overlay'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_https_url(self, get_git_config_val):
        get_git_config_val.return_value = \
            'https://JNRowe@github.com/JNRowe/misc-overlay.git'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_https_url_no_suffix(self, get_git_config_val):
        get_git_config_val.return_value = \
            'https://JNRowe@github.com/JNRowe/misc-overlay'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_http_url(self, get_git_config_val):
        get_git_config_val.return_value = \
            'http://JNRowe@github.com/JNRowe/misc-overlay.git'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_http_url_no_suffix(self, get_git_config_val):
        get_git_config_val.return_value = \
            'http://JNRowe@github.com/JNRowe/misc-overlay'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_http_url_no_auth(self, get_git_config_val):
        get_git_config_val.return_value = \
            'http://github.com/JNRowe/misc-overlay.git'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    def test_http_url_no_suffix_no_auth(self, get_git_config_val):
        get_git_config_val.return_value = \
            'http://github.com/JNRowe/misc-overlay'
        assert_equals(utils.get_repo(), 'JNRowe/misc-overlay')

    @patch('hubugs.utils.get_git_config_val')
    @raises(ValueError)
    def test_broken_url(self, get_git_config_val):
        get_git_config_val.return_value = 'git://github.com/misc-overlay.git'
        utils.get_repo()

    @patch('hubugs.utils.get_git_config_val')
    @raises(ValueError)
    def test_no_url(self, get_git_config_val):
        get_git_config_val.return_value = None
        utils.get_repo()

    @patch('hubugs.utils.get_git_config_val')
    @raises(ValueError)
    def test_invalid_url(self, get_git_config_val):
        get_git_config_val.return_value = 'http://example.com/dog.git'
        utils.get_repo()



#
# coding=utf-8
"""models - GitHub models for hubugs"""
# Copyright (C) 2010-2012  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import micromodels


class ISODateTimeField(micromodels.DateTimeField):

    """ISO-8601 compliant date and time field."""

    def __init__(self, **kwargs):
        super(ISODateTimeField, self).__init__(format='%Y-%m-%dT%H:%M:%SZ',
                                               **kwargs)


class User(micromodels.Model):

    """GitHub user model."""

    id = micromodels.IntegerField()
    avatar_url = micromodels.CharField()
    gravatar_id = micromodels.CharField()
    login = micromodels.CharField()
    url = micromodels.CharField()

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.login)


class Label(micromodels.Model):

    """GitHub issue label model."""

    color = micromodels.CharField()
    name = micromodels.CharField()
    url = micromodels.CharField()

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.name)


class PullRequest(micromodels.Model):

    """GitHub issue pull request data model."""

    diff_url = micromodels.CharField()
    html_url = micromodels.CharField()
    patch_url = micromodels.CharField()

    def __nonzero__(self):
        return bool(self.patch_url)

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.patch_url)


class Issue(micromodels.Model):

    """GitHub issue model."""

    id = micromodels.IntegerField()
    assignee = micromodels.ModelField(User)
    body = micromodels.CharField()
    body_html = micromodels.CharField()
    closed_at = ISODateTimeField()
    closed_by = micromodels.ModelField(User)
    comments = micromodels.IntegerField()
    created_at = ISODateTimeField()
    html_url = micromodels.CharField()
    labels = micromodels.ModelCollectionField(Label)
    milestone = micromodels.CharField()
    number = micromodels.IntegerField()
    pull_request = micromodels.ModelField(PullRequest)
    state = micromodels.CharField()
    title = micromodels.CharField()
    updated_at = ISODateTimeField()
    url = micromodels.CharField()
    user = micromodels.ModelField(User)

    def __repr__(self):
        return "<%s %s %r>" % (self.__class__.__name__, self.id,
                               self.title[:20])


class Comment(micromodels.Model):

    """GitHub issue comment model."""

    id = micromodels.IntegerField()
    body = micromodels.CharField()
    body_html = micromodels.CharField()
    created_at = ISODateTimeField()
    updated_at = ISODateTimeField()
    url = micromodels.CharField()
    user = micromodels.ModelField(User)

    def __repr__(self):
        return "<%s %s %r>" % (self.__class__.__name__, self.id,
                               self.body[:20])


class Application(micromodels.Model):

    """GitHub OAuth application model."""

    name = micromodels.CharField()
    url = micromodels.CharField()

    def __repr__(self):
        return "<%s %r>" % (self.__class__.__name__, self.name)


class Authorisation(micromodels.Model):

    """GitHub OAuth authorisation model."""

    id = micromodels.IntegerField()
    app = micromodels.ModelField(Application)
    created_at = ISODateTimeField()
    note = micromodels.CharField()
    note_url = micromodels.CharField()
    scopes = micromodels.FieldCollectionField(micromodels.CharField())
    token = micromodels.CharField()
    updated_at = ISODateTimeField()
    url = micromodels.CharField()

    def __repr__(self):
        return "<%s %s %r>" % (self.__class__.__name__, self.id,
                               self.note[:20])


class Organization(micromodels.Model):

    """GitHub issue organisation model."""

    id = micromodels.IntegerField()
    avatar_url = micromodels.CharField()
    gravatar_id = micromodels.CharField()
    login = micromodels.CharField()
    type = micromodels.CharField()
    url = micromodels.CharField()


class Repository(micromodels.Model):

    """GitHub repository model."""

    id = micromodels.IntegerField()
    clone_url = micromodels.CharField()
    created_at = ISODateTimeField()
    description = micromodels.CharField()
    fork = micromodels.BooleanField()
    forks = micromodels.IntegerField()
    git_url = micromodels.CharField()
    has_downloads = micromodels.BooleanField()
    has_issues = micromodels.BooleanField()
    has_wiki = micromodels.BooleanField()
    homepage = micromodels.CharField()
    html_url = micromodels.CharField()
    language = micromodels.CharField()
    master_branch = micromodels.CharField()
    mirror_url = micromodels.CharField()
    name = micromodels.CharField()
    open_issues = micromodels.IntegerField()
    owner = micromodels.ModelField(User)
    private = micromodels.BooleanField()
    pushed_at = ISODateTimeField()
    size = micromodels.IntegerField()
    url = micromodels.CharField()
    watchers = micromodels.IntegerField()



#
# coding=utf-8
"""hubugs - Simple client for GitHub issues"""
# Copyright (C) 2010-2012  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import _version


__version__ = _version.dotted
__date__ = _version.date
__author__ = "James Rowe <jnrowe@gmail.com>"
__copyright__ = "Copyright (C) 2010-2012  James Rowe <jnrowe@gmail.com>"
__license__ = "GNU General Public License Version 3"
__credits__ = ""
__history__ = "See git repository"

from email.utils import parseaddr

__doc__ += """.

``hubugs`` is a very simple client for working with `GitHub's issue tracker`_.

.. _GitHub's issue tracker: http://github.com/blog/411-github-issue-tracker

.. moduleauthor:: `%s <mailto:%s>`__
""" % parseaddr(__author__)

# This is here to workaround UserWarning messages caused by path fiddling in
# dependencies
try:
    import pkg_resources  # NOQA
except ImportError:
    pass

import atexit
import errno
import getpass
import logging
import os
# Used by raw_input, when imported
import readline  # NOQA
import sys
import webbrowser

from base64 import b64encode

import argh
import httplib2


logging.basicConfig(level=logging.ERROR,
                    format="%(asctime)s - %(message)s",
                    datefmt="%Y-%m-%dT%H:%M:%S")
atexit.register(logging.shutdown)


from . import (template, utils)


COMMANDS = []


def command(func):
    """Simple decorator to add function to ``COMMANDS`` list.

    The purpose of this decorator is to make the definition of commands simpler
    by reducing duplication, it is purely a convenience.

    :param func func: Function to wrap
    :rtype: ``func``
    :returns: Original function

    """
    COMMANDS.append(func)
    return func


# Convenience wrappers for defining command arguments
# pylint: disable-msg=C0103
bugs_arg = argh.arg("bugs", nargs="+", type=int,
                    help="bug number(s) to operate on")

message_arg = argh.arg("-m", "--message", help="comment text")

order_arg = argh.arg("-o", "--order", default="number",
                     choices=["number", "updated"],
                     help="sort order for listing bugs")

states_arg = argh.arg("-s", "--state", default="open",
                      choices=["open", "closed", "all"],
                      help="state of bugs to operate on")

stdin_arg = argh.arg("--stdin", default=False,
                     help="read message from standard input")

title_arg = argh.arg("title", help="title for the new bug", nargs="?")
body_arg = argh.arg("body", help="body for the new bug", nargs="?")

label_add_arg = argh.arg("-a", "--add", action="append", default=[],
                         help="add label to issue", metavar="label")
label_create_arg = argh.arg("-c", "--create", action="append", default=[],
                            help="create new label and add to issue",
                            metavar="label")
label_remove_arg = argh.arg("-r", "--remove", action="append", default=[],
                            help="remove label from issue", metavar="label")
# pylint: enable-msg=C0103


@command
@argh.arg('--local', default=False,
          help='set access token for local repository only')
def setup(args):
    """setup GitHub access token"""
    if not utils.SYSTEM_CERTS:
        yield utils.warn('Falling back on bundled certificates')
    if utils.CURL_CERTS:
        yield utils.warn('Using certs specified in $CURL_CERTS')
    default_user = os.getenv("GITHUB_USER",
                             utils.get_git_config_val("github.user",
                                                      getpass.getuser()))
    try:
        user = raw_input('GitHub user? [%s] ' % default_user)
        if not user:
            user = default_user
        password = getpass.getpass('GitHub password? ')
    except KeyboardInterrupt:
        return
    private = argh.confirm('Support private repositories', default=True)
    data = {
        'scopes': ['repo' if private else 'public_repo'],
        'note': 'hubugs',
        'note_url': 'https://github.com/JNRowe/hubugs'
    }

    # Unfortunately, we need to forcibly define the header to workaround
    # GitHub sending a 404(in an effort to stop information leakage)
    header = {
        'Authorization': 'Basic ' + b64encode(":".join([user, password]))
    }
    r, auth = args.req_post('https://api.github.com/authorizations', body=data,
                            headers=header, model='Authorisation')
    utils.set_git_config_val('hubugs.token', auth.token, args.local)
    yield utils.success('Configuration complete!')


@command
@argh.alias("list")
@states_arg
@argh.arg("-l", "--label", help="list bugs with specified label",
          metavar="label")
@order_arg
def list_bugs(args):
    """listing bugs"""
    bugs = []
    params = {}
    if args.label:
        params['labels'] = args.label

    states = ["open", "closed"] if args.state == "all" else [args.state, ]
    for state in states:
        _params = params.copy()
        _params['state'] = state
        r, _bugs = args.req_get('', params=_params, model=['Issue', ])
        bugs.extend(_bugs)

    yield template.display_bugs(bugs, args.order, state=args.state,
                                project=args.repo_obj)


@command
@argh.arg("-f", "--full", default=False, help="show bug including comments")
@argh.arg("-p", "--patch", default=False,
          help="display patches for pull requests")
@argh.arg("-o", "--patch-only", default=False,
          help="display only the patch content of pull requests")
@argh.arg("-b", "--browse", default=False, help="open bug in web browser")
@bugs_arg
def show(args):
    """displaying bugs"""
    tmpl = template.get_template('view', '/issue.txt')
    for bug_no in args.bugs:
        if args.browse:
            webbrowser.open_new_tab("https://github.com/%s/issues/%d"
                                    % (args.project, bug_no))
            continue
        r, bug = args.req_get(bug_no, model='Issue')

        if args.full and bug.comments:
            r, comments = args.req_get('%s/comments' % bug_no,
                                       model=['Comment', ])
        else:
            comments = []
        if (args.patch or args.patch_only) and bug.pull_request:
            r, c = args.req_get(bug.pull_request.patch_url, is_json=False)
            patch = c.decode('utf-8')
        else:
            patch = None
        yield tmpl.render(bug=bug, comments=comments, full=True,
                          patch=patch, patch_only=args.patch_only,
                          project=args.repo_obj)


@command
@argh.alias("open")
@label_add_arg
@label_create_arg
@stdin_arg
@title_arg
@body_arg
@argh.wrap_errors(template.EmptyMessageError)
def open_bug(args):
    """opening new bugs"""
    utils.sync_labels(args)
    if args.stdin:
        text = sys.stdin.readlines()
    elif not args.title:
        text = template.edit_text("open").splitlines()
    if args.stdin or not args.title:
        title = text[0]
        body = "\n".join(text[1:])
    else:
        title = args.title
        body = args.body
    data = {'title': title, 'body': body, 'labels': args.add + args.create}
    r, bug = args.req_post('', body=data, model='Issue')
    yield utils.success("Bug %d opened" % bug.number)


@command
@stdin_arg
@message_arg
@bugs_arg
@argh.wrap_errors(template.EmptyMessageError)
def comment(args):
    """commenting on bugs"""
    if args.stdin:
        message = sys.stdin.read()
    elif args.message:
        message = args.message
    else:
        message = template.edit_text()
    for bug in args.bugs:
        args.req_post('%s/comments' % bug, body={'body': message})


@command
@stdin_arg
@title_arg
@body_arg
@bugs_arg
@argh.wrap_errors(template.EmptyMessageError)
def edit(args):
    """editing bugs"""
    if (args.title or args.stdin) and len(args.bugs) > 1:
        raise argh.CommandError("Can not use --stdin or command line "
                                "title/body with multiple bugs")
    for bug in args.bugs:
        if args.stdin:
            text = sys.stdin.readlines()
        elif not args.title:
            r, current = args.req_get(bug, model='Issue')
            current_data = {"title": current.title, "body": current.body}
            text = template.edit_text("open", current_data).splitlines()
        if args.stdin or not args.title:
            title = text[0]
            body = "\n".join(text[1:])
        else:
            title = args.title
            body = args.body

        data = {'title': title, 'body': body}
        args.req_post(bug, body=data)


@command
@stdin_arg
@message_arg
@bugs_arg
def close(args):
    """closing bugs"""
    if args.stdin:
        message = sys.stdin.read()
    elif not args.message:
        try:
            message = template.edit_text()
        except template.EmptyMessageError:
            # Message isn't required for closing, but it is good practice
            message = None
    else:
        message = args.message
    for bug in args.bugs:
        if message:
            args.req_post('%s/comments' % bug, body={'body': message})
        args.req_post(bug, body={'state': 'closed'})


@command
@stdin_arg
@message_arg
@bugs_arg
def reopen(args):
    """reopening closed bugs"""
    if args.stdin:
        message = sys.stdin.read()
    elif not args.message:
        try:
            message = template.edit_text()
        except template.EmptyMessageError:
            # Message isn't required for reopening, but it is good practice
            message = None
    else:
        message = args.message
    for bug in args.bugs:
        if message:
            args.req_post('%s/comments' % bug, body={'body': message})
        args.req_post(bug, body={'state': 'open'})


@command
@label_add_arg
@label_create_arg
@label_remove_arg
@argh.arg("-l", "--list", default=False, help="list available labels")
@argh.arg("bugs", nargs="*", type=int,
          help="bug number(s) to operate on")
def label(args):
    """labelling bugs"""
    label_names = utils.sync_labels(args)

    if args.list:
        print ", ".join(sorted(label_names))
        return

    for bug_no in args.bugs:
        r, bug = args.req_get(bug_no, model='Issue')
        labels = [label.name for label in bug.labels]
        labels.extend(args.add + args.create)

        for string in args.remove:
            labels.remove(string)
        args.req_post(bug_no, body={'labels': labels})


@command
@argh.wrap_errors(template.EmptyMessageError)
def report_bug(args):
    """report a new bug against hubugs"""
    local = args.project == 'JNRowe/hubugs'
    args.project = 'JNRowe/hubugs'

    import blessings, html2text, jinja2, micromodels, pygments  # NOQA
    versions = dict([(m.__name__, getattr(m, '__version__', 'No version info'))
                     for m in argh, blessings, html2text, httplib2, jinja2,
                        micromodels, pygments])
    data = {
        'local': local,
        'sys': sys,
        'version': _version.dotted,
        'versions': versions,
        'certs': utils.CA_CERTS,
    }
    text = template.edit_text("hubugs_report", data).splitlines()
    title = text[0]
    body = "\n".join(text[1:])

    data = {'title': title, 'body': body, 'labels': args.add + args.create}
    r, bug = args.req_post('', body=data, model='Issue')
    yield utils.success("Bug %d opened against hubugs, thanks!" % bug.number)


def main():
    """Main command-line entry point.

    :rtype: ``int``
    :return: Exit code

    """
    description = __doc__.splitlines()[0].split("-", 1)[1]
    epilog = "Please report bugs to the JNRowe/hubugs project"
    parser = argh.ArghParser(description=description, epilog=epilog,
                             version="%%(prog)s %s" % __version__)
    parser.add_argument("-p", "--project", action=utils.ProjectAction,
                        help="GitHub project to operate on",
                        metavar="project")
    parser.add_argument("-u", "--host-url", default='https://api.github.com',
                        help="GitHub Enterprise host to connect to",
                        metavar="url")
    parser.add_commands(COMMANDS)
    try:
        parser.dispatch(pre_call=utils.setup_environment)
    except (utils.RepoError, EnvironmentError, ValueError) as error:
        print utils.fail(error.message)
        return errno.EINVAL
    except httplib2.ServerNotFoundError:
        print utils.fail("Project lookup failed.  Network or GitHub down?")
        return errno.ENXIO

if __name__ == '__main__':
    sys.exit(main())



from collections import namedtuple
from unittest import TestCase

from datetime import (datetime, timedelta)
from mock import patch
from nose.tools import (assert_equals, assert_true, raises)
from pygments import (formatters, lexers)

from hubugs import template


# We only test forced styling output of blessings, as blessings handles the
# sys.stdout.isatty() flipping
template.utils.T = template.utils.blessings.Terminal(force_styling=True)


class Colourise(TestCase):
    def test_color(self):
        assert_equals(template.colourise('s', 'red'),
                      u'\x1b[38;5;1ms\x1b[m\x1b(B')

    def test_background_color(self):
        assert_equals(template.colourise('s', 'on blue'),
                                         u'\x1b[48;5;4ms\x1b[m\x1b(B')

    def test_attribute(self):
        assert_equals(template.colourise('s', 'bold'), u'\x1b[1ms\x1b[m\x1b(B')

    @raises(TypeError)
    def test_invalid_colour(self):
        assert_equals(template.colourise('s', 'mauve with a hint of green'),
                                         's')


class Highlight(TestCase):
    def pyg_side_effect(*args, **kwargs):
        return namedtuple('Call', 'args kwargs')(args, kwargs)

    @patch('hubugs.template.pyg_highlight')
    def test_highlight(self, pyg_highlight):
        pyg_highlight.side_effect = self.pyg_side_effect
        result = template.highlight('+++ a\n--- b\n+Test\n')
        assert_true(isinstance(result.args[2], lexers.DiffLexer))
        assert_true(isinstance(result.args[3],
                               formatters.terminal.TerminalFormatter))

    @patch('hubugs.template.pyg_highlight')
    def test_highlight_lexer(self, pyg_highlight):
        pyg_highlight.side_effect = self.pyg_side_effect
        result = template.highlight('True', 'python')
        assert_true(isinstance(result.args[2], lexers.PythonLexer))

    @patch('hubugs.template.pyg_highlight')
    def test_highlight_formatter(self, pyg_highlight):
        pyg_highlight.side_effect = self.pyg_side_effect
        result = template.highlight('True', formatter='terminal256')
        assert_true(isinstance(result.args[3],
                               formatters.terminal256.Terminal256Formatter))


class EditText(TestCase):
    @staticmethod
    def getmtime_side_effect(file, response=range(100)):
        return response.pop()

    @staticmethod
    def check_call_side_effect(args):
        open(args[1], 'a').write('Some message')

    @patch('subprocess.check_call')
    @raises(template.EmptyMessageError)
    def test_no_message(self, check_call):
        check_call.return_value = True
        template.edit_text()

    @patch('subprocess.check_call')
    @patch('os.path.getmtime')
    def test_message(self, getmtime, check_call):
        getmtime.side_effect = self.getmtime_side_effect
        check_call.side_effect = self.check_call_side_effect

        assert_equals(template.edit_text(), 'Some message')

    @patch('subprocess.check_call')
    @patch('os.path.getmtime')
    def test_message_comments(self, getmtime, check_call):
        getmtime.side_effect = self.getmtime_side_effect
        check_call.side_effect = self.check_call_side_effect

        assert_equals(template.edit_text(), 'Some message')

    @patch('subprocess.check_call')
    @raises(template.EmptyMessageError)
    def test_message_prefill(self, check_call):
        check_call.return_value = True
        assert_equals(template.edit_text('open',
                                         data={'title': 'Some message'}),
                      'Some message')


class Html2Text(TestCase):
    def test_basic(self):
        assert_equals(template.html2text('<h3>hello</h3>'), '### hello')

    def test_width(self):
        para = """<p>This is a long paragraph that needs wrapping to work so it
        doesn't make you want to claw your eyes out."""
        assert_equals(template.html2text(para).count('\n'), 1)
        assert_equals(template.html2text(para, width=20).count('\n'), 5)


class RelativeTime(TestCase):
    def test_last_year(self):
        dt = datetime.utcnow() - timedelta(days=365)
        assert_equals(template.relative_time(dt), 'last year')

    def test_months_ago(self):
        dt = datetime.utcnow() - timedelta(days=70)
        assert_equals(template.relative_time(dt), 'about two months ago')

    def test_month_ago(self):
        dt = datetime.utcnow() - timedelta(days=30)
        assert_equals(template.relative_time(dt), 'last month')

    def test_weeks_ago(self):
        dt = datetime.utcnow() - timedelta(days=21)
        assert_equals(template.relative_time(dt), 'about three weeks ago')

    def test_days_ago(self):
        dt = datetime.utcnow() - timedelta(days=4)
        assert_equals(template.relative_time(dt), 'about four days ago')

    def test_yesterday(self):
        dt = datetime.utcnow() - timedelta(days=1)
        assert_equals(template.relative_time(dt), 'yesterday')

    def test_hours_ago(self):
        dt = datetime.utcnow() - timedelta(hours=5)
        assert_equals(template.relative_time(dt), 'about five hours ago')

    def test_hour_ago(self):
        dt = datetime.utcnow() - timedelta(hours=1)
        assert_equals(template.relative_time(dt), 'about an hour ago')

    def test_minutes_ago(self):
        dt = datetime.utcnow() - timedelta(minutes=6)
        assert_equals(template.relative_time(dt), 'about six minutes ago')

    def test_seconds_ago(self):
        dt = datetime.utcnow() - timedelta(seconds=12)
        assert_equals(template.relative_time(dt), 'about 12 seconds ago')



#
# coding=utf-8
"""utils - Utility functions for hubugs"""
# Copyright (C) 2010-2012  James Rowe <jnrowe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import json
import os
import re
import subprocess
import sys
import urllib

from functools import partial

import argh
import blessings
import httplib2

from . import (_version, models)


T = blessings.Terminal()

SYSTEM_CERTS = \
    not httplib2.CA_CERTS.startswith(os.path.dirname(httplib2.__file__))
CURL_CERTS = False
if not SYSTEM_CERTS and sys.platform.startswith('linux'):
    for cert_file in ['/etc/ssl/certs/ca-certificates.crt',
                      '/etc/pki/tls/certs/ca-bundle.crt']:
        if os.path.exists(cert_file):
            CA_CERTS = cert_file
            SYSTEM_CERTS = True
            break
elif not SYSTEM_CERTS and sys.platform.startswith('freebsd'):
    if os.path.exists('/usr/local/share/certs/ca-root-nss.crt'):
        CA_CERTS = '/usr/local/share/certs/ca-root-nss.crt'
        SYSTEM_CERTS = True
elif os.path.exists(os.getenv('CURL_CA_BUNDLE', '')):
    CA_CERTS = os.getenv('CURL_CA_BUNDLE')
    CURL_CERTS = True
if not SYSTEM_CERTS and not CURL_CERTS:
    CA_CERTS = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                            "GitHub_certs.crt")


# Set up informational message functions
def _colourise(text, colour):
    """Colour text, if possible.

    :param str text: Text to colourise
    :param str colour: Colour to display text in
    :rtype: ``str``
    :return: Colourised text, if possible

    """
    return getattr(T, colour.replace(' ', '_'))(text)


def success(text):
    """Output a success message.

    :param str text:  Text to format
    :rtype: ``str`
    :return: Bright green text, if possible

    """
    return _colourise(text, 'bright green')


def fail(text):
    """Output a failure message.

    :param str text:  Text to format
    :rtype: ``str`
    :return: Bright red text, if possible

    """
    return _colourise(text, 'bright red')


def warn(text):
    """Output a warning message.

    :param str text:  Text to format
    :rtype: ``str`
    :return: Bright yellow text, if possible

    """
    return _colourise(text, 'bright yellow')


class HttpClientError(ValueError):

    """Error raised for client error status codes."""

    def __init__(self, message, response, content):
        super(HttpClientError, self).__init__(message)
        self.response = response
        self.content = content


class RepoError(ValueError):

    """Error raised for invalid repository values."""


class ProjectAction(argh.utils.argparse.Action):

    """argparse action class for setting project."""

    def __call__(self, parser, namespace, project, option_string=None):
        """Set fully qualified GitHub project name."""
        if not "/" in project:
            user = os.getenv("GITHUB_USER", get_git_config_val("github.user"))
            if not user:
                raise parser.error("No GitHub user setting!")
            project = "%s/%s" % (user, project)
        namespace.project = project


def check_output(args, **kwargs):
    """Simple check_output implementation for Python 2.6 compatibility.

    :param list args: Command and arguments to call
    :rtype: ``str``
    :return: Command output
    :raise subprocess.CalledProcessError: If command execution fails

    """
    try:
        return subprocess.check_output(args, **kwargs)
    except AttributeError:
        process = subprocess.Popen(args, stdout=subprocess.PIPE, **kwargs)
        output, unused_err = process.communicate()
        retcode = process.poll()
        if retcode:
            raise subprocess.CalledProcessError(retcode, args[0])
        return output


def get_github_api():
    """Create a GitHub API instance.

    :rtype: ``httplib2.Http``
    :return: GitHub HTTP session

    """
    if sys.platform == 'darwin':
        user_cache_dir = os.path.expanduser("~/Library/Caches")
    else:
        user_cache_dir = os.path.join(os.getenv("HOME", "/"), ".cache")
    xdg_cache_dir = os.getenv("XDG_CACHE_HOME")
    cache_dir = os.path.join(xdg_cache_dir or user_cache_dir, "hubugs")

    return httplib2.Http(cache_dir, ca_certs=CA_CERTS)


def get_git_config_val(key, default=None, local_only=False):
    """Fetch a git configuration value.

    :param str key: Configuration value to fetch
    :param str default: Default value to use, if key isn't set
    :param bool local_only: Fetch configuration values from repo config only
    :rtype: ``str``
    :return: Git config value, if set

    """
    cmd = ['git', 'config', ]
    if local_only:
        cmd.append('--local')
    cmd.extend(['--get', key])
    try:
        output = check_output(cmd).strip()
    except subprocess.CalledProcessError:
        output = default
    return output


def set_git_config_val(key, value, local_only=False):
    """Set a git configuration value.

    :param str key: Configuration value to fetch
    :param str value: Value to set
    :param bool local_only: Set configuration values from repo config only

    """
    cmd = ['git', 'config', ]
    if not local_only:
        cmd.append('--global')
    cmd.extend([key, value])
    try:
        check_output(cmd, stderr=subprocess.STDOUT).strip()
    except subprocess.CalledProcessError as e:
        raise argh.CommandError(e.output)


def get_editor():
    """Choose a suitable editor.

    This follows the method defined in :manpage:`git-var(1)`

    :rtype: ``str``
    :return: Users chosen editor, or ``vi`` if not set

    """
    # Match git for editor preferences
    editor = os.getenv("GIT_EDITOR")
    if not editor:
        editor = get_git_config_val("core.editor")
        if not editor:
            editor = os.getenv("VISUAL", os.getenv("EDITOR", "vi"))
    return editor


def get_repo():
    """Extract GitHub project name from git config.

    :rtype: ``str``
    :return: GitHub project name, including user

    """
    data = get_git_config_val("remote.origin.url", local_only=True)
    if not data:
        raise RepoError("No `origin' remote found")
    match = re.match(r"""
        (?:git(?:@|://)  # SSH or git protocol
          |https?://
           (?:.*@)?)  # HTTP URLs support optional auth data
        github.com[:/]  # hostname, : sep for SSH URL
        ([^/]+/.*?)  # project
        (?:.git)?$  # .git suffix is optional in all clone URLs""",
                     data, re.VERBOSE)
    if match:
        return match.groups()[0]
    else:
        raise RepoError("Unknown project, specify with `--project' option")


def setup_environment(args):
    """Configure execution environment for commands dispatch.

    :param argparse.Namespace args: argparse namespace to operate on

    """
    if not args.project:
        args.project = get_repo()

    command = args.function.__name__

    http = get_github_api()

    HEADERS = {
        'Accept': 'application/vnd.github.beta.full+json',
        'User-Agent': 'hubugs/%s' % _version.dotted,
    }

    # We use manual auth when calling setup
    use_auth = not command == 'setup'
    if use_auth:
        token = os.getenv("HUBUGS_TOKEN", get_git_config_val("hubugs.token"))
        if not token:
            raise EnvironmentError("No hubugs authorisation token found!  "
                                   "Run 'hubugs setup' to create a token")
        HEADERS["Authorization"] = "token %s" % token

    def http_method(url, method='GET', params=None, body=None, headers=None,
                    model=None, is_json=True):
        if headers:
            headers.update(HEADERS)
        else:
            headers = HEADERS
        if not isinstance(url, basestring) or not url.startswith('http'):
            url = '%s/repos/%s/issues%s%s' % (args.host_url, args.project,
                                              '/' if url else '', url)
        if params:
            url += '?' + urllib.urlencode(params)
        if is_json and body:
            body = json.dumps(body)
        r, c = http.request(url, method=method, body=body, headers=headers)
        if is_json:
            c = json.loads(c)
        if str(r.status)[0] == '4':
            raise HttpClientError(str(r.status), r, c)
        if model:
            if isinstance(model, list):
                creator = getattr(models, model[0]).from_dict
                c = [creator(d) for d in c]
            else:
                c = getattr(models, model).from_dict(c)
        return r, c

    args.req_get = http_method
    args.req_post = partial(http_method, method='POST')

    # Make the repository information available, if it will be useful
    # Note: We skip this step for `show -b' for speed, see #20
    if not command == 'setup' \
        and not (command == 'show' and args.browse == True):
        try:
            r, c = args.req_get('%s/repos/%s' % (args.host_url, args.project))
        except HttpClientError as e:
            if e.content['message'] == 'Not Found':
                raise RepoError('Invalid project %r' % args.project)
            raise
        args.repo_obj = models.Repository.from_dict(c)
        if not args.repo_obj.has_issues:
            raise RepoError("Issues aren't enabled for %r" % args.project)


def sync_labels(args):
    """Manage labels for a project.

    :param argparse.Namespace args: argparse namespace to operate on
    :rtype: ``list``
    :return: List of project's label names

    """
    labels_url = '%s/repos/%s/labels' % (args.host_url, args.project)
    r, c = args.req_get(labels_url)
    label_names = [label['name'] for label in c]

    for label in args.add:
        if label not in label_names:
            raise ValueError('No such label %r' % label)
    for label in args.create:
        if label in label_names:
            print warn('%r label already exists' % label)
        else:
            data = {'name': label, 'color': '000000'}
            args.req_post(labels_url, body=data)
    return label_names + args.add + args.create



#! /usr/bin/python -tt

import sys

from setuptools import setup

# Hack to import _version file without importing hubugs/__init__.py, its
# purpose is to allow import without requiring dependencies at this point.
_version = {}
execfile('hubugs/_version.py', {}, _version)

install_requires = ['argh', 'blessings', 'html2text', 'httplib2>=0.7.4',
                    'Jinja2>=2', 'micromodels', 'Pygments']
if sys.version_info[:2] < (2, 7):
    install_requires.append('argparse')


setup(
    name='hubugs',
    version=_version['dotted'],
    description="Simple client for GitHub issues",
    long_description=open("README.rst").read(),
    author="James Rowe",
    author_email="jnrowe@gmail.com",
    url="https://github.com/JNRowe/hubugs",
    license="GPL-3",
    keywords="github bugs cli",
    packages=['hubugs', ],
    include_package_data=True,
    package_data={'': ['templates/*/*.mkd', "templates/*/*.txt", '*.crt'], },
    entry_points={'console_scripts': ['hubugs = hubugs:main', ]},
    install_requires=install_requires,
    zip_safe=False,
    test_suite="nose.collector",
    tests_require=['nose', 'mock'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Topic :: Communications",
        "Topic :: Internet",
        "Topic :: Software Development",
        "Topic :: Software Development :: Bug Tracking",
        "Topic :: Terminals",
        "Topic :: Utilities",
    ],
)
